import os
import requests
import datetime
import json
import base64
import logging
import time
from web3.auto import w3
from web3 import Web3

MOCK_JWT = 'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJib3QtaWQiOiIweDEzazM4N2IzNzc2OWNlMjQyMzZjNDAzZTc2ZmMzMGYwMWZhNzc0MTc2ZTE0MTZjODYxeWZlNmMwN2RmZWY3MWYiLCJleHAiOjE2NjAxMTk0NDMsImlhdCI6MTY2MDExOTQxMywianRpIjoicWtkNWNmYWQtMTg4NC0xMWVkLWE1YzktMDI0MjBhNjM5MzA4IiwibmJmIjoxNjYwMTE5MzgzLCJzdWIiOiIweDU1NmY4QkU0MmY3NmMwMUY5NjBmMzJDQjE5MzZEMmUwZTBFYjNGNEQifQ.9v5OiiYhDoEbhZ-abbiSXa5y-nQXa104YCN_2mK7SP0'


def fetch_jwt(claims={}, expiresAt=None) -> str:
    if(os.environ['NODE_ENV'] != 'production'):
        return MOCK_JWT

    host_name = 'forta-jwt-provider'
    port = 8515
    path = '/create'
    uri = 'http://{host_name}:{port}{path}'.format(
        host_name=host_name, port=port, path=path)

    if((expiresAt != None) and (isinstance(expiresAt, datetime.datetime) == False)):
        raise Exception("expireAt must be of type datetime")

    if expiresAt is not None:
        exp_in_sec = expiresAt.timestamp()
        claims["exp"] = exp_in_sec

    response = requests.request("POST", uri, json={'claims': claims})

    if response.status_code == 200:
        data = response.json()
        return data.get('token')
    else:
        raise Exception(
            "Error occured with response fetching jwt token.", response)


def verify_jwt(token: str, polygonUrl: str = 'https://polygon-rpc.com') -> bool:
    splitJwt = token.split('.')
    rawHeader = splitJwt[0]
    rawPayload = splitJwt[1]

    header = json.loads(base64.urlsafe_b64decode(
        rawHeader + '==').decode('utf-8'))
    payload = json.loads(base64.urlsafe_b64decode(
        rawPayload + '==').decode('utf-8'))

    alg = header['alg']
    botId = payload['bot-id']
    expiresAt = payload['exp']
    signerAddress = payload['sub']

    if (signerAddress is None) or (botId is None):
        logging.warning('Invalid claim')
        return False

    if alg != 'ETH':
        logging.warning('Unexpected signing method: {alg}'.format(alg=alg))
        return False

    currentUnixTime = time.mktime(datetime.datetime.utcnow().utctimetuple())

    if expiresAt < currentUnixTime:
        logging.warning('Jwt expired')
        return False

    msg = '{header}.{payload}'.format(header=rawHeader, payload=rawPayload)
    msgHash = w3.keccak(text=msg)
    b64signature = splitJwt[2]
    signature = base64.urlsafe_b64decode(f'{b64signature}=').hex()
    recoveredSignerAddress = w3.eth.account.recoverHash(
        msgHash, signature=signature)

    if recoveredSignerAddress != signerAddress:
        logging.warn('Signature invalid: expected={signerAddress}, got={recoveredSignerAddress}'.format(
            signerAddress=signerAddress, recoveredSignerAddress=recoveredSignerAddress))
        return False

    w3Client = Web3(Web3.HTTPProvider(polygonUrl))
    DISPTACHER_ABI = [{"inputs": [{"internalType": "uint256", "name": "agentId", "type": "uint256"}, {"internalType": "uint256", "name": "scannerId", "type": "uint256"}],
                       "name": "areTheyLinked", "outputs": [{"internalType": "bool", "name": "", "type": "bool"}], "stateMutability": "view", "type": "function"}]
    # Source: https://docs.forta.network/en/latest/smart-contracts/
    DISPATCH_CONTRACT = "0xd46832F3f8EA8bDEFe5316696c0364F01b31a573"
    contract = w3Client.eth.contract(
        address=DISPATCH_CONTRACT, abi=DISPTACHER_ABI)

    areTheyLinked = contract.functions.areTheyLinked(
        int(botId, 0), int(recoveredSignerAddress, 0)).call()

    return areTheyLinked


class DecodedJwt:
    def __init__(self, dict):
        self.header = dict.get('header')
        self.payload = dict.get('payload')


def decode_jwt(token):
    # Adding need 4 byte for pythons b64decode
    header = base64.urlsafe_b64decode(
        token.split('.')[0] + '==').decode('utf-8')
    payload = base64.urlsafe_b64decode(
        token.split('.')[1] + '==').decode('utf-8')

    return DecodedJwt({
        "header": header,
        "payload": payload
    })



from .utils import get_alerts


def test_get_alerts_returns_more_than_one_alert():
    assert True
#     response = get_alerts(
#         {'bot_ids': ["0x79af4d8e0ea9bd28ed971f0c54bcfe2e1ba0e6de39c4f3d35726b15843990a51"]})
#     assert len(response.alerts) > 0


def test_get_alerts_parse_alert():
    assert True
    #     response = get_alerts(
    #         {'bot_ids': ["0x79af4d8e0ea9bd28ed971f0c54bcfe2e1ba0e6de39c4f3d35726b15843990a51"]})
    #     alert = response.alerts[0]
    #     print(alert.description)
    #     assert True


def test_get_alerts_exception_on_400_response():
    assert True
    #     try:
    #         get_alerts({'bot_ids': ["0x79af4d8e0ea9bd28ed971f0c54bcfe2e1ba0e6de39c4f3d35726b15843990a51"],
    #                    'created_since': '2022-11-02T14:15:09.783203651Z'})
    #         assert False

    #     except Exception as err:
    #         assert err



from .jwt import MOCK_JWT, fetch_jwt
import responses
import os
import pytest


@responses.activate
def test_return_mock_jwt():
    # Register error response because we should not make a real call
    rsp1 = responses.Response(
        url="/",
        method="POST",
        json={"error": "not found"},
        status=404,
    )
    responses.add(rsp1)

    os.environ['NODE_ENV'] = 'dev'

    token = fetch_jwt()

    assert token == MOCK_JWT


@responses.activate
def test_return_valid_JWT():
    testJWT = "testJWT"
    # Register response because we should not make a real call
    rsp1 = responses.Response(
        url="http://forta-jwt-provider:8515/create",
        method="POST",
        json={"token": testJWT},
        status=200,
    )
    responses.add(rsp1)

    os.environ['NODE_ENV'] = 'production'

    token = fetch_jwt()

    assert token == testJWT


@responses.activate
def test_JWT_should_throw_exception():
    with pytest.raises(Exception) as e_info:

        # Register error response because we should not make a real call
        rsp1 = responses.Response(
            url="http://forta-jwt-provider:8515/create",
            method="POST",
            json={"message": 'Bad request'},
            status=400,
        )
        responses.add(rsp1)

        os.environ['NODE_ENV'] = 'production'

        token = fetch_jwt()



import sys
import os
from jsonc_parser.parser import JsoncParser
import requests
from web3.auto import w3
from web3 import Web3
from .forta_graphql import AlertsResponse


def get_web3_provider():
    from . import web3Provider
    return web3Provider


def get_forta_config():
    config = {}
    # try to read global config
    global_config_path = os.path.join(
        os.path.expanduser('~'), '.forta', 'forta.config.json')
    if os.path.isfile(global_config_path):
        global_config = JsoncParser.parse_file(global_config_path)
        config = {**config, **global_config}
    # try to read local project config
    config_flag_index = sys.argv.index(
        '--config') if '--config' in sys.argv else -1
    local_config_file = None if config_flag_index == - \
        1 else sys.argv[config_flag_index + 1]
    local_config_path = os.path.join(
        os.getcwd(), local_config_file if local_config_file else 'forta.config.json')
    if os.path.isfile(local_config_path):
        local_config = JsoncParser.parse_file(local_config_path)
        config = {**config, **local_config}
    return config


def get_json_rpc_url():
    if 'JSON_RPC_HOST' in os.environ:
        return f'http://{os.environ["JSON_RPC_HOST"]}{":"+os.environ["JSON_RPC_PORT"] if "JSON_RPC_PORT" in os.environ else ""}'

    config = get_forta_config()
    if "jsonRpcUrl" not in config:
        return "https://cloudflare-eth.com/"
    if not str(config.get("jsonRpcUrl")).startswith("http"):
        raise Exception("jsonRpcUrl must begin with http(s)")
    return config.get("jsonRpcUrl")


def get_transaction_receipt(tx_hash):
    from .receipt import Receipt  # avoid circular import
    web3_provider = get_web3_provider()
    receipt = web3_provider.eth.get_transaction_receipt(tx_hash)
    return Receipt({
        "status": receipt.get("status") == 1,
        "root": receipt.get("root"),
        "gas_used": receipt.get("gasUsed"),
        "cumulative_gas_used": receipt.get("cumulativeGasUsed"),
        "logs_bloom": receipt.get("logsBloom").hex(),
        "contract_address": None if receipt.get("contractAddress") == None else receipt.get("contractAddress").lower(),
        "block_number": receipt.get("blockNumber"),
        "block_hash": receipt.get("blockHash").hex(),
        "transaction_index": receipt.get("transactionIndex"),
        "transaction_hash": receipt.get("transactionHash").hex(),
        "logs": list(map(lambda log: {
            "address": log.get("address").lower(),
            "topics": list(map(lambda topic: topic.hex(), log.get("topics", []))),
            "data": log.get("data"),
            "log_index": log.get("logIndex"),
            "block_number": log.get("blockNumber"),
            "block_hash": log.get("blockHash").hex(),
            "transaction_index": log.get("transactionIndex"),
            "transaction_hash": log.get("transactionHash").hex(),
            "removed": log.get("removed")
        }, receipt.get("logs", []))),
    })


def create_block_event(dict):
    from .block_event import BlockEvent  # avoid circular import
    return BlockEvent(dict)


def create_transaction_event(dict):
    from .transaction_event import TransactionEvent  # avoid circular import
    return TransactionEvent(dict)


def create_alert_event(dict):
    from .alert_event import AlertEvent  # avoid circular import
    return AlertEvent(dict)


def get_alerts(dict):
    from .forta_graphql import AlertQueryOptions
    forta_api = "https://api.forta.network/graphql"
    headers = {"content-type": "application/json"}
    query_options = AlertQueryOptions(dict)
    payload = query_options.get_query()

    response = requests.request(
        "POST", forta_api, json=payload, headers=headers)

    if response.status_code == 200:
        data = response.json().get('data')

        if data:
            return AlertsResponse(data.get('alerts'))
    else:
        message = response.text
        raise Exception(message)


def assert_non_empty_string_in_dict(dict, key):
    assert_key_in_dict(dict, key)
    assert isinstance(dict[key], str) and len(
        dict[key]) > 0, f'{key} must be non-empty string'


def assert_enum_value_in_dict(dict, key, enum):
    assert_key_in_dict(dict, key)
    assert isinstance(dict[key], enum), f'{key} must be valid enum value'


def assert_key_in_dict(dict, key):
    assert key in dict, f'{key} is required'


def hex_to_int(strVal):
    if not strVal or type(strVal) == int:
        return strVal
    return int(strVal, 16) if type(strVal) == str and strVal.startswith('0x') else int(strVal, 10)


def keccak256(val):
    return Web3.keccak(text=val).hex()



class Block:
    def __init__(self, dict):
        self.difficulty = dict.get('difficulty')
        self.extra_data = dict.get('extraData', dict.get('extra_data'))
        self.gas_limit = dict.get('gasLimit', dict.get('gas_limit'))
        self.gas_used = dict.get('gasUsed', dict.get('gas_used'))
        self.hash = dict.get('hash')
        self.logs_bloom = dict.get('logsBloom', dict.get('logs_bloom'))
        self.miner = dict.get('miner')
        self.mix_hash = dict.get('mixHash', dict.get('mix_hash'))
        self.nonce = dict.get('nonce')
        self.number = dict.get('number')
        self.parent_hash = dict.get('parentHash', dict.get('parent_hash'))
        self.receipts_root = dict.get(
            'receiptsRoot', dict.get('receipts_root'))
        self.sha3_uncles = dict.get('sha3Uncles', dict.get('sha3_uncles'))
        self.size = dict.get('size')
        self.state_root = dict.get('stateRoot', dict.get('state_root'))
        self.timestamp = dict.get('timestamp')
        self.total_difficulty = dict.get(
            'totalDifficulty', dict.get('total_difficulty'))
        self.transactions = dict.get('transactions')
        self.transactions_root = dict.get(
            'transactionsRoot', dict.get('transactions_root'))
        self.uncles = dict.get('uncles')



from .utils import hex_to_int


class Transaction:
    def __init__(self, dict):
        self.hash = dict.get('hash')
        self.from_ = dict.get('from')
        self.to = dict.get('to')
        self.nonce = dict.get('nonce')
        self.gas = hex_to_int(dict.get('gas'))
        self.gas_price = hex_to_int(
            dict.get('gasPrice', dict.get('gas_price')))
        self.value = hex_to_int(dict.get('value'))
        self.data = dict.get('data')
        self.r = dict.get('r')
        self.s = dict.get('s')
        self.v = dict.get('v')



from .alert import Alert


class AlertEvent:
    def __init__(self, dict):
        self.alert = Alert(dict.get('alert', {}))

    @property
    def alert_id(self):
        return self.alert.alert_id

    @property
    def name(self):
        return self.alert.name

    @property
    def hash(self):
        return self.alert_hash

    @property
    def alert_hash(self):
        return self.alert.hash

    @property
    def bot_id(self):
        return self.alert.source.bot.id

    @property
    def transaction_hash(self):
        return self.alert.source.transaction_hash

    @property
    def block_hash(self):
        return self.alert.source.block.hash

    @property
    def block_number(self):
        return self.alert.source.block.number

    @property
    def chain_id(self):
        return self.alert.chain_id



class Alert:
    def __init__(self, dict):
        from .label import Label
        self.addresses = dict.get('addresses')
        self.alert_id = dict.get('alertId')
        self.contracts = list(map(lambda t: Contract(t), dict.get(
            'contracts', []))) if dict.get('contracts') is not None else []
        self.created_at = dict.get('createdAt')
        self.description = dict.get('description')
        self.finding_type = dict.get('findingType')
        self.name = dict.get('name')
        self.hash = dict.get('hash')
        self.protocol = dict.get('protocol')
        self.severity = dict.get('severity')
        self.source = Source(dict.get('source')) if dict.get(
            'source') is not None else None
        self.metadata = dict.get('metadata')
        self.projects = list(map(lambda t: Project(t), dict.get(
            'projects', []))) if dict.get('projects') is not None else []
        self.scan_node_count = dict.get('scanNodeCount')
        self.alert_document_type = dict.get('alertDocumentType')
        self.related_alerts = dict.get('relatedAlerts')
        self.chain_id = dict.get('chainId')
        self.labels = list(map(lambda t: Label(t), dict.get(
            'labels', []))) if dict.get('labels') is not None else []


class Source:
    def __init__(self, dict):
        self.transaction_hash = dict.get('transactionHash')
        self.block = SourceBlock(dict.get('block')) if dict.get(
            'block') is not None else None
        self.bot = SourceBot(dict.get('bot')) if dict.get(
            'bot') is not None else None
        self.source_alert = SourceAlert(dict.get('sourceAlert')) if dict.get(
            'sourceAlert') is not None else None


class SourceBlock:
    def __init__(self, dict):
        self.timestamp = dict.get('timestamp')
        self.chain_id = dict.get('chainId')
        self.hash = dict.get('hash')
        self.number = dict.get('number')


class SourceBot:
    def __init__(self, dict):
        self.id = dict.get('id')
        self.reference = dict.get('reference')
        self.image = dict.get('image')


class SourceAlert:
    def __init__(self, dict):
        self.hash = dict.get('hash')
        self.bot_id = dict.get('botId')
        self.timestamp = dict.get('timestamp')
        self.chain_id = dict.get('chainId')


class Contract:
    def __init__(self, dict):
        self.address = dict.get('address')
        self.name = dict.get('name')
        self.project_id = dict.get('projectId')


class Project:
    def __init__(self, dict):
        self.id = dict.get('id')
        self.name = dict.get('name')
        self.contacts = ProjectContacts(dict.get('contacts')) if dict.get(
            'contacts') is not None else None
        self.website = dict.get('website')
        self.token = ProjectToken(dict.get('token')) if dict.get(
            'token') is not None else None
        self.social = ProjectSocial(dict.get('social')) if dict.get(
            'social') is not None else None


class ProjectContacts:
    def __init__(self, dict):
        self.security_email_address = dict.get('securityEmailAddress')
        self.general_email_address = dict.get('generalEmailAddress')


class ProjectToken:
    def __init__(self, dict):
        self.symbol = dict.get('symbol')
        self.name = dict.get('name')
        self.decimals = dict.get('decimals')
        self.chain_id = dict.get('chainId')
        self.address = dict.get('address')


class ProjectSocial:
    def __init__(self, dict):
        self.twitter = dict.get('twitter')
        self.github = dict.get('github')
        self.everest = dict.get('everest')
        self.coingecko = dict.get('coingecko')



from .alert import Alert

class AlertCursor:
    def __init__(self, dict):
        self.alert_id = dict.get('alertId')
        self.block_number = dict.get('blockNumber')


class AlertQueryOptions:
    def __init__(self, dict):
        self.bots = dict.get('bot_ids')
        self.addresses = dict.get('addresses')
        self.alertHash = dict.get('alert_hash')
        self.alertId = dict.get('alert_id')
        self.alertIds = dict.get('alert_ids')
        self.chainId = dict.get('chain_id')
        self.createdSince = dict.get('created_since')
        self.createdBefore = dict.get('created_before')
        self.first = dict.get('first')
        self.after = dict.get('starting_cursor')
        self.projectId = dict.get('project_id')
        self.scanNodeConfirmations = dict.get('scan_node_confirmations')
        self.severities = dict.get('severities')
        self.transactionHash = dict.get('transaction_hash')
        self.blockSortDirection = dict.get('block_sort_direction')
        self.blockDateRange = dict.get('block_date_range')
        self.blockNumberRange = dict.get('block_number_range')

    def get_query(self):
        query = """
          query($input: AlertsInput) {
            alerts(input: $input) {
                alerts {
                    alertId
                    addresses
                    contracts {
                        address
                        name
                        projectId
                    }
                    createdAt
                    description
                    hash
                    metadata
                    name
                    projects {
                        id
                    }
                    protocol
                    scanNodeCount
                    severity
                    source {
                        transactionHash
                        bot {
                            chainIds
                            createdAt
                            description
                            developer
                            docReference
                            enabled
                            id
                            image
                            name
                            reference
                            repository
                            projects
                            scanNodes
                            version
                        }
                        block {
                            number
                            hash
                            timestamp
                            chainId
                        }
                        sourceAlert {
                            hash
                            botId
                            timestamp
                            chainId
                        }
                    }
                    alertDocumentType
                    findingType
                    relatedAlerts
                    chainId
                    labels {
                        label
                        confidence
                        entity
                        entityType
                        remove
                    }
                }
                pageInfo {
                    hasNextPage
                    endCursor {
                        blockNumber
                        alertId
                    }
                }
            }
          }
          """
        query_variables = vars(self)
        filtered_query_variables = {k:v for k,v in query_variables.items() if v is not None}
        return dict(query=query, variables={"input": filtered_query_variables})


class AlertsResponse:
    def __init__(self, dict):
        self.alerts = list(map(lambda t: Alert(t), dict.get('alerts', []))) if dict.get('alerts') is not None else []
        self.page_info = PageInfo(dict.get('pageInfo')) if dict.get('pageInfo') is not None else None

class PageInfo:
    def __init__(self,dict):
        self.has_next_page = dict.get('hasNextPage')
        self.end_cursor = AlertCursor(dict.get('endCursor')) if dict.get('endCursor') is not None else None


import json
from hexbytes import HexBytes
from .event_type import EventType
from .network import Network
from .transaction import Transaction
from .receipt import Log
from .trace import Trace


class TxEventBlock:
    def __init__(self, dict):
        self.hash = dict.get('hash')
        self.number = dict.get('number')
        self.timestamp = dict.get('timestamp')


class TransactionEvent:
    def __init__(self, dict):
        typeVal = dict.get('type', "BLOCK")
        self.type = EventType[typeVal] if type(
            typeVal) == str else EventType(typeVal)
        networkVal = dict.get('network', "MAINNET")
        networkVal = int(networkVal) if type(
            networkVal) == str and networkVal.isdigit() else networkVal
        self.network = Network[networkVal] if type(
            networkVal) == str else Network(networkVal)
        self.transaction = Transaction(dict.get('transaction', {}))
        self.traces = list(map(lambda t: Trace(t), dict.get('traces', [])))
        self.addresses = dict.get('addresses', {})
        self.block = TxEventBlock(dict.get('block', {}))
        self.logs = list(map(lambda l: Log(l), dict.get('logs', [])))
        self.contract_address = dict.get(
            'contractAddress', dict.get('contract_address'))

    @property
    def hash(self):
        return self.transaction.hash

    @property
    def to(self):
        return self.transaction.to

    @property
    def from_(self):
        return self.transaction.from_

    @property
    def gas_price(self):
        return self.transaction.gas_price

    @property
    def timestamp(self):
        return self.block.timestamp

    @property
    def block_number(self):
        return self.block.number

    @property
    def block_hash(self):
        return self.block.hash

    def filter_log(self, abi, contract_address=''):
        abi = abi if isinstance(abi, list) else [abi]
        abi = [json.loads(abi_item) for abi_item in abi]
        logs = self.logs
        # filter logs by contract address, if provided
        if (contract_address):
            contract_address = contract_address if isinstance(
                contract_address, list) else [contract_address]
            contract_address_map = {
                address.lower(): True for address in contract_address}
            logs = filter(lambda log: log.address.lower()
                          in contract_address_map, logs)
        # determine which event names to filter
        event_names = []
        for abi_item in abi:
            if abi_item['type'] == "event":
                event_names.append(abi_item['name'])
        # parse logs
        results = []
        from . import web3Provider
        contract = web3Provider.eth.contract(
            "0x0000000000000000000000000000000000000000", abi=abi)
        for log in logs:
            log.topics = [HexBytes(topic) for topic in log.topics]
            for event_name in event_names:
                try:
                    results.append(
                        contract.events[event_name]().processLog(log))
                except:
                    continue  # TODO see if theres a better way to handle 'no matching event' error
        return results

    def filter_function(self, abi, contract_address=''):
        abi = abi if isinstance(abi, list) else [abi]
        abi = [json.loads(abi_object) for abi_object in abi]
        # determine where to look for function calls (i.e. transaction object or traces)
        sources = [{'data': self.transaction.data, 'to': self.transaction.to}]
        if self.traces:
            sources = [{'data': trace.action.input,
                        'to': trace.action.to} for trace in self.traces]
        # filter by contract address, if provided
        if (contract_address):
            contract_address = contract_address if isinstance(
                contract_address, list) else [contract_address]
            contract_address_map = {
                address.lower(): True for address in contract_address}
            sources = filter(
                lambda source: source['to'] and source['to'].lower() in contract_address_map, sources)
        # parse function inputs
        results = []
        from . import web3Provider
        contract = web3Provider.eth.contract(
            "0x0000000000000000000000000000000000000000", abi=abi)
        for source in sources:
            try:
                decoded_function = contract.decode_function_input(
                    source['data'])
                decoded_function[1]["address"] = source['to'].lower()
                results.append(decoded_function)
            except:
                continue  # TODO see if theres a better way to handle 'no matching function' error
        return results



from enum import IntEnum
from .utils import assert_enum_value_in_dict, assert_non_empty_string_in_dict


class EntityType(IntEnum):
    Unknown = 0
    Address = 1
    Transaction = 2
    Block = 3
    Url = 4


class Label:
    def __init__(self, dict):
        entityTypeVal = dict.get('entity_type', dict.get('entityType'))
        self.entity_type = EntityType[entityTypeVal.title()] if type(
            entityTypeVal) == str else EntityType(entityTypeVal)
        assert_enum_value_in_dict(self.__dict__, 'entity_type', EntityType)
        assert_non_empty_string_in_dict(dict, 'entity')
        self.entity = dict['entity']
        self.confidence = dict['confidence']
        self.label = dict['label']
        self.remove = dict.get('remove', False)

    def toDict(self):
        d = dict(self.__dict__, **{
            'entityType': self.entity_type,
        })
        return {k: v for k, v in d.items() if v is not None}
